@@ -56,8 +56,8 @@ Value *polly::createLoop(Value *LB, Value *UB, Value *Stride,
                          PollyIRBuilder &Builder, LoopInfo &LI,
                          DominatorTree &DT, BasicBlock *&ExitBB,
                          ICmpInst::Predicate Predicate,
-                         ScopAnnotator *Annotator, bool Parallel,
-                         bool UseGuard) {
+                         ScopAnnotator *Annotator, bool Parallel, bool UseGuard,
+                         bool LoopVectDisabled) {
   Function *F = Builder.GetInsertBlock()->getParent();
   LLVMContext &Context = F->getContext();
 
@@ -74,7 +74,7 @@ Value *polly::createLoop(Value *LB, Value *UB, Value *Stride,
 
   // Update LoopInfo
   Loop *OuterLoop = LI.getLoopFor(BeforeBB);
-  Loop *NewLoop = new Loop();
+  Loop *NewLoop = LI.AllocateLoop();
 
   if (OuterLoop)
     OuterLoop->addChildLoop(NewLoop);
@@ -132,7 +132,7 @@ Value *polly::createLoop(Value *LB, Value *UB, Value *Stride,
   // Create the loop latch and annotate it as such.
   BranchInst *B = Builder.CreateCondBr(LoopCondition, HeaderBB, ExitBB);
   if (Annotator)
-    Annotator->annotateLoopLatch(B, NewLoop, Parallel);
+    Annotator->annotateLoopLatch(B, NewLoop, Parallel, LoopVectDisabled);
 
   IV->addIncoming(IncrementedIV, HeaderBB);
   if (GuardBB)
